import numpy as np
from scipy.optimize import brenth

class Solver :
    """
    Numerical solver. The class allows to move step by step. The main method here is "step". 
    All other methods are auxiliary functions for "step". Note that throughout the code, 
    a lot of meaningless quantities are computed. These are precomputations to increase efficiency.    
    """
    def __init__(self,
                 B = 250,
                 S = 1,
                 n = 0.8,
                 lx = 1,
                 nx = 150,
                 cfl = 0.006,
                 change_cfl = 500,
                 coeff_speed = 2,
                 coeff_diff = 20
                 ):
            """
            Parameters
            ----------
            B : float, optional
                The default is 250.
            S : float, optional
                The default is 1.
            n : int or float, optional
                Value of the HB law power. If 1, must be an int. 
                The default is 0.8.
            lx : int or float, optional
                Length of the box. 
                The default is 1.
            nx : int, optional
                Number of points of discretization. 
                The default is 150.
            cfl : float, optional
                The CFL condition to use for early steps. Shouldn't be changed. 
                The default is 0.006.
            change_cfl : int, optional
                Number of iterations after which switching from the previous CFL condition to a numerical 
                stability condition based on the characteristic speed and diffusion of the equation. 
                The default is 500.
            coeff_speed : int or float, optional
                Coefficient by which to divide the speed stability condition criteria. 
                The default is 2.
            coeff_diff : int or float, optional
                Coefficient by which to divide the diffusion stability condition criteria. 
                The default is 20.
    
            Returns
            -------
            None.    
            """
            # initialisation - physical parameters
            
            self.B = B   # "Bingham number", see paper
            self.S = S   # slope param
            self.n = n   # power law param
            self.n2 = 2*n+1
            self.n2_ = 2*n
            self.n1 = 1+n
            self.n21 = 1/((2*n+1)*(n+1))
            self.n21_1 = n * self.n21
            self.n_2 = 2*n*n
            self.n_1 = 1/n
            self.n1_1 = 1/n+1
            
            self.lx = lx                 # length of the physical domain
            
            # initialisation - discretisation parameters
            self.nx = nx                 # number of mesh points (boundaries included)
            self.nx1 = nx-1
            self.nx2 = nx-2
            self.dx = lx/(nx-1)          # space step Delta x
            self.dx_2 = self.dx * self.dx
            self.dx2 = 1/(2 * self.dx)
            self.dx_1 = (nx-1)/lx
            self.Sdx = self.S*self.dx
            self.cfl = cfl               # "CFL number"
            self.change_cfl = change_cfl # number of iterations after which the cfl evolves
            
            # Convention hk, with k for time iterator
            # i is for space index: i=1 left boundary; i=nx right wall
            self.hk = np.zeros(nx) # initial condition for h
            
            self.dt = cfl * self.dx
            self.dtadv = self.dt
            self.dtdif = self.dt
            self.coeff_speed = coeff_speed
            self.coeff_diff = coeff_diff
            
            # flux
            self.q = np.zeros(nx)
            self.q[0] = 1
            
            # step number
            self.k = 0
            
            # time
            self.t = 0
            
            # a test to see if we reached some terminal condition
            self.end = False
            

    def fsolqhb(self, x, h):
        """
        Computes the discrete flow rate at the entrance of the box.

        Parameters
        ----------
        x : float
            Height at position 0 (= i).
        h : float
            Height at position 1 (= i).

        Returns
        -------
        float
            Flow rate.
        """
        Shxd = self.S -(h-x)*self.dx_1
        if (x * Shxd <= self.B) :
            return -1
        else :
            tmp = x - self.B / Shxd
            return ((Shxd*tmp)**self.n_1) * tmp * (self.n2*x - self.n*tmp)*self.n21_1-1
    
    def optim(self, h):
        """
        Given the height of the point of position 1, computes and updates the height of the 
        point of position 0 by use of a root searching algorithm.

        Parameters
        ----------
        h : float
            Height at position 1 (= i).

        Returns
        -------
        None.
        """
        hSdx = (h-self.Sdx)*0.5
        start = hSdx+np.sqrt(hSdx*hSdx+self.B*self.dx)
        end = start + 0.005
        while self.fsolqhb(end,h) < 0:
            end += 0.05
        self.hk[0] = brenth((lambda u: self.fsolqhb(u, h)), start, end)

    def step(self): 
      """
      Solves one time step.

      Returns
      -------
      None.
      """
      # compute gradient of hk
      hk = self.hk[1:self.nx1]
      Shkx = self.S- (self.hk[2:] - self.hk[:self.nx2])*self.dx2 # never use hkx alone
      delta = np.sign(Shkx)
      Shkx = np.abs(Shkx)
      Shkx_1 = 1/Shkx
      # compute the flow rate q
      Y = hk - self.B / Shkx
      Y[Y<0] = 0
      YnShkxn = (Y * Shkx) ** self.n_1
      self.q[1:self.nx1] = delta * YnShkxn * Y * (self.n2*hk - self.n*Y) * self.n21_1
    
      # numerical stability condition
      if self.k >= self.change_cfl :

        nBShkx = self.n*self.B*Shkx_1
        diffus = self.n21 * YnShkxn * Shkx_1 * (2*hk*nBShkx + self.n1*hk+2*(nBShkx*nBShkx))
        diffcoeff = np.max(diffus)
        speed = YnShkxn * hk
        spidcoef = np.max(speed)
        self.dtadv = self.dx/(self.coeff_speed*spidcoef)
        self.dtdif = self.dx_2/(self.coeff_diff*diffcoeff)
        self.dt = min(self.dtadv,self.dtdif)
    
      # update the new h
      self.hk[1:] -= (self.dt*self.dx_1)*(self.q[1:]-self.q[:self.nx1]) # if q>0 everywhere
      # update h[0]
      self.optim(self.hk[1])
      self.k += 1
      self.t += self.dt
