import numpy as np
from scipy.optimize import brenth

class SolverB :
    """
    Numerical solver specific for the case n=1. The class is similar to the class 
    "Solve" of the module HB. The user should refer to it for the usage of the 
    common method, because they are the same, only replacing n by 1.    
    """
    
    def __init__(self,
                 B = 250,
                 S = 1,
                 lx = 1,
                 nx = 150,
                 cfl = 0.006,
                 change_cfl = 500,
                 coeff_speed = 2,
                 coeff_diff = 20
                 ):
        
            # initialisation - physical parameters
            self.B = B   # "Bingham number", see paper
            self.S = S   # slope param
            self.n2 = 3
            self.n2_ = 2
            self.n1 = 2
            self.n21 = 1/6
            self.n21_1 = 1/6
            self.n_2 = 2
            self.n_1 = 1
            self.n1_1 = 2
            
            self.lx = lx                 # length of the physical domain
            
            # initialisation - discretisation parameters
            self.nx = nx                 # number of mesh points (boundaries included)
            self.nx1 = nx-1
            self.nx2 = nx-2
            self.dx = lx/(nx-1)          # space step Delta x
            self.dx_2 = self.dx * self.dx
            self.dx2 = 1/(2 * self.dx)
            self.dx_1 = (nx-1)/lx
            self.Sdx = self.S*self.dx
            self.cfl = cfl               # "CFL number"
            self.change_cfl = change_cfl # number of iterations after which the cfl evolves
            
            # Convention hk, with k for time iterator
            # i is for space index: i=1 left boundary; i=nx right wall
            self.hk = np.zeros(nx) # initial condition for h
            
            self.dt = cfl * self.dx
            self.dtadv = self.dt
            self.dtdif = self.dt
            self.coeff_speed = coeff_speed
            self.coeff_diff = coeff_diff
            
            # flux
            self.q = np.zeros(nx)
            self.q[0] = 1
            
            # step number
            self.k = 0
            
            # time
            self.t = 0
            
            # a test to see if we reached some terminal condition
            self.end = False
            
            # precomputations for the polynomial used in the calculation of h(0)
            self._7_ = 2*self.dx_1
            self._6_ = 8*S
            self._6_h = -8*self.dx_1
            self._5_ = 12*S*S*self.dx-3*B
            self._5_h = -24*S
            self._5_hh = 12*self.dx_1
            self._4_ = -9*B*S*self.dx+8*S*S*S*self.dx*self.dx
            self._4_h = 9*B-24*S*S*self.dx
            self._4_hh = 24*S
            self._4_hhh = -8*self.dx_1
            self._3_ = -9*B*S*S*self.dx*self.dx+2*S*S*S*S*self.dx*self.dx*self.dx-6
            self._3_h = 18*B*S*self.dx-8*S*S*S*self.dx*self.dx
            self._3_hh = 12*S*S*self.dx-9*B
            self._3_hhh = -8*S
            self._3_hhhh = self._7_
            self._2_ = -3*B*S*S*S*self.dx*self.dx*self.dx-18*S*self.dx
            self._2_h = 9*B*S*S*self.dx*self.dx+18
            self._2_hh = -9*B*S*self.dx
            self._2_hhh = 3*B
            self._1_ = B*B*B*self.dx*self.dx-18*S*S*self.dx*self.dx
            self._1_h = 36*S*self.dx
            self._1_hh = -18
            self._0_ = B*B*B*S*self.dx*self.dx*self.dx-6*S*S*S*self.dx*self.dx*self.dx
            self._0_h = -B*B*B*self.dx*self.dx+18*S*S*self.dx*self.dx
            self._0_hh = -18*S*self.dx
            self._0_hhh = 6
            self.pol = np.zeros(8)
            self.pol[7] = self._7_

    def fsolqhb(self, x, h):
        Shxd = self.S -(h-x)*self.dx_1
        if (x * Shxd <= self.B) :
            return -1
        else :
            tmp = x - self.B / Shxd
            return Shxd*tmp * tmp * (3*x - tmp)/6-1
    
    def optim(self, h):
        hSdx = (h-self.Sdx)*0.5
        start = hSdx+np.sqrt(hSdx*hSdx+self.B*self.dx)
        end = start + 0.005
        while self.fsolqhb(end,h) < 0:
            end += 0.05
            #print("aug")
        self.hk[0] = brenth((lambda u: self.fsolqhb(u, h)), start, end)

    def step(self): # one time step       
      # compute gradient of hk
      hk = self.hk[1:self.nx1]
      Shkx = self.S- (self.hk[2:] - self.hk[:self.nx2])*self.dx2 # never use hkx alone
      delta = np.sign(Shkx)
      Shkx = np.abs(Shkx)
      Shkx_1 = 1/Shkx
      # compute the flux, q, only bulk needed
      Y = hk - self.B / Shkx
      Y[Y<0] = 0
      YnShkxn = Y * Shkx
      self.q[1:self.nx1] = delta * YnShkxn * Y * (3*hk - Y)/6
    
      if self.k >= self.change_cfl :

        nBShkx = self.B*Shkx_1
        diffus = YnShkxn * Shkx_1 * (2*hk*nBShkx + 2*hk+2*(nBShkx*nBShkx))/6
        diffcoeff = np.max(diffus)
        speed = YnShkxn * hk
        spidcoef = np.max(speed)
        self.dtadv = self.dx/(self.coeff_speed*spidcoef)
        self.dtdif = self.dx_2/(self.coeff_diff*diffcoeff)
        self.dt = min(self.dtadv,self.dtdif)
    
      # update the new h - bulk points
      self.hk[1:] -= (self.dt*self.dx_1)*(self.q[1:]-self.q[:self.nx1]) # if q>0 everywhere
      # update the new h - the left B.C. + non-linear solve on q
      self.optim(self.hk[1])
      self.k += 1
      self.t += self.dt
      
