import os
import time
import numpy as np
repo_path = os.environ["ZENODO_REPO_CPB"]
os.chdir(repo_path)
from src.solvers.HB import Solver
from src.solvers.bingham import SolverB
from src.utils.data_base import update_couples_grid, extract_all_couples

def gen(BS, n=1, nx=301, lx=1, num_front_max=50, prefix=repo_path+"data/PDE_solutions/", 
        grid_name=None, check=True, avoid=None):
    """
    Generator of solutions for the database. For all couples, "num_front_max" height profiles are 
    saved, evenly distributed in space. The end is the wall touch. Can optionally update a grid 
    descriptor while computing. For each snapshot, the height profile, the number of iterations, 
    the time, dtadv and dtdif are stored. The duration of the computation as well as some parameters 
    are stored in the header of the file. The output file is of the form:
        n_1_B_50.0_S_60.0

    Parameters
    ----------
    BS : list
        List of couples to compute.
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. 
        The default is 1.
    nx : int, optional
        Number of points of discretization. 
        The default is 301.
    lx : int or float, optional
        Length of the box. 
        The default is 1.
    num_front_max : int, optional
        Number of profiles to save. The default is 51.
    prefix : str, optional
        Directory of the database. 
        The default is "res/data_base/".
    grid_name : str or NoneType, optional
        The name of the grid to update. If None, no grid is updated.
        The default is None.
    check : bool, optional
        If a grid is updated, argument to decide whether or not the algorithm looks for redundancies
        before updating. The default is True.
    avoid : NoneType or list, optional
        List of couples that shouldn't be computed. If None, these are all already computed couples.
        The default is None.

    Returns
    -------
    None.
    """
    # update avoid
    if type(avoid) == type(None) :
        avoid = extract_all_couples(n=n, array=False)
    # spatial flags for save
    fracL = 1.0/num_front_max
    fraclx = fracL*lx
    x = np.linspace(0, lx, num=nx)
    
    for B, S in BS :
        suffix = "n_{}_B_{}_S_{}".format(n, B, S)
        path = prefix+suffix
        if not([B,S] in avoid) :
            # initialisation
            python = -np.ones((num_front_max, nx+4))
            print("Computing n={}, B={}, S={}".format(n, B, S))
            if n==1 :
                sol = SolverB(B=B, S=S, nx=nx, coeff_diff=2)
                print("Using Bingham specific algorithm")
            else :
                sol = Solver(B=B, S=S, nx=nx, n=n, coeff_diff=2)
            xtarget = fraclx
            num_front = 0
            start = time.time()
            while not(sol.end) :
                sol.step()
                ind = sol.hk.argmin() # look for the position of the front of the solution
                if sol.hk[-1] == 0 :
                    xfront = x[ind]
                else :
                    sol.end = True # if we touch the wall, that's the end
                    xfront = xtarget+1
                if xfront > xtarget :
                    python[num_front, :nx] = sol.hk
                    python[num_front, nx] = sol.k
                    python[num_front, nx+1] = sol.t
                    python[num_front, nx+2] = sol.dtadv
                    python[num_front, nx+3] = sol.dtdif
                    xtarget += fraclx
                    num_front += 1
                    print("advancement {:.3f}%, duration {:.3f}s".format((num_front/num_front_max)*100,\
                                                                         time.time()-start))
            end = time.time()
            if -1. in python :
                i = python[:,0].argmin() # take off some possibly unused lines
                python = python[:i]
            else :
                i = python.shape[0]
            header = "parameters : nx {} n {} lx {} cfl {} change_cfl".format(nx, n, sol.lx, sol.cfl)+\
               " {} coeff_speed {} ".format(sol.change_cfl, sol.coeff_speed)+\
                "coeff_diff {} number of time saved {} tmax {} ".format(sol.coeff_diff, i, sol.t)+\
                    "duration {} number of iterations {} the {} first".format(end-start, sol.k, nx)+\
                        "elt are h, then k, then t, then dtadv then dtdif for each row"
            np.savetxt(path, python, header=header)
            if grid_name != None :
                update_couples_grid([(B,S)], grid_name, check=check)
        else :
            print(suffix+" already exists")


def regular_grid_wrapper_1(n=1,
                           bmin=0.5,
                           bmax=250,
                           smin=0.05,
                           smax=120,
                           num_s=20,
                           num_b=20):
    """
    Add a regular grid in the database. Values of B and S are rounded at two digits.

    Parameters
    ----------
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. The default is 1.
    bmin : float, optional
        Minimal value of B. The default is 0.5.
    bmax : float, optional
        Maximal value of B. The default is 250.
    smin : float, optional
        Minimal value of S. The default is 0.05.
    smax : float, optional
        Maximal value of S. The default is 120.
    num_s : int, optional
        Number of values of S. The default is 20.
    num_b : int, optional
        Number of values of B. The default is 20.

    Returns
    -------
    None.
    """
    list_B = np.linspace(bmin, bmax, num=num_b)
    for i in range(num_b) :
        list_B[i] = int(list_B[i] * 100)/100.
    list_S = np.linspace(smin, smax, num=num_s)
    for i in range(num_s) :
        list_S[i] = int(list_S[i] * 100)/100.
    list_S = np.flip(list_S)
    BS = np.array(np.meshgrid(list_B, list_S)).T.reshape(-1, 2)
    gen(BS, n=n)
    name = "regular_grid_1_n_{}".format(n)
    update_couples_grid(BS, name)

def gen_from_grid_file(file, n=1):
    """
    Compute the couples of the given file. The file should be readable as a 2D array, one line per 
    couple.

    Parameters
    ----------
    file : str
        File where the couples are listed.
    n : int or float, optional
        Value of the HB law power. If 1, must be an int.
        The default is 1.

    Returns
    -------
    None.
    """
    couples = np.loadtxt(file)
    avoid = extract_all_couples(n=n, array=False)
    gen(couples, avoid=avoid, n=n)
