import os
import numpy as np
import matplotlib.pyplot as plt
repo_path = os.environ["ZENODO_REPO_CPB"]
os.chdir(repo_path)
from src.metamodel.PCE import PCE
from src.inversion.nelder_mead import Nelder_Mead
import src.inversion.utils as u

def inverse(profile, algo=None, meta=None,**args):
    """
    Use the metamodel+Nelder-Mead procedure to solve an inverse problem. Both
    the Nelder-Mead and PCE object can be instantiated within this function.

    Parameters
    ----------
    profile : numpy array
        Height profile of the inverse problem.
    algo : Nelder_Mead or NoneType, optional
        The Nelder_Mead object to perform the inversion. If None, it is
        instantiated within the function using the provided metamodel (see below).
        Arguments for the initialisation of the Nelder_Mead object can be given
        via the **args.
        The default is None.
    meta : PCE or NoneType, optional
        Metamodel to use to instantiate the Nelder_Mead object if it is not
        already provided. If None, it is instantiated within the function. 
        Arguments for its initialisation can be given via the **args.
        The default is None.
    **args :
        Arguments for algo and meta.

    Returns
    -------
    numpy array
        The (B,S) couple estimated by Nelder-Mead.
    """
    if algo == None :
        if meta == None :
            meta = PCE(*args)
        algo = Nelder_Mead(meta,*args)
    return algo.inverse(profile)

def inverse_batch(profiles, algo=None, meta=None,**args):
    """
    Use the metamodel+Nelder-Mead procedure to solve multiple inverse problems. 
    Both the Nelder-Mead and PCE object can be instantiated within this function.

    Parameters
    ----------
    profiles : numpy array
        Height profiles of the inverse problem. Shape should be (num_profiles,301).
    algo : Nelder_Mead or NoneType, optional
        The Nelder_Mead object to perform the inversion. If None, it is
        instantiated within the function using the provided metamodel (see below).
        Arguments for the initialisation of the Nelder_Mead object can be given
        via the **args.
        The default is None.
    meta : PCE or NoneType, optional
        Metamodel to use to instantiate the Nelder_Mead object if it is not
        already provided. If None, it is instantiated within the function. 
        Arguments for its initialisation can be given via the **args.
        The default is None.
    **args :
        Arguments for algo and meta.

    Returns
    -------
    res : numpy array
        The (B,S) couples estimated by Nelder-Mead. Shape is (num_profiles,2).
    """
    if algo == None :
        if meta == None :
            meta = PCE(*args)
        algo = Nelder_Mead(meta,*args)
    num_profiles = profiles.shape[0]
    res = np.zeros((num_profiles,2))
    for i in range(num_profiles):
        res[i] = algo.inverse(profiles[i])
    return res

def comparison_exp_inversion(filename="illustrative_profile",path=repo_path+"data/experiments/",plot=True,do_nondim=True):
    """
    Use the metamodel+Nelder-Mead procedure to solve an inverse problem and display
    the comparison between the experimental data and the fitted profile.

    Parameters
    ----------
    filename : str, optional
        Name of the file. 
        The default is "illustrative_profile".
    path : str, optional
        Path to the file. 
        The default is repo_path+"data/experiments/".
    plot : bool, optional
        Whether to display the comparison or not. 
        The default is True.
    do_nondim : bool, optional
        Should be set to True if the profile is not nondimensional, in which
        case it will be rescaled. To do so, the physical dimensions of the 
        corresponding experiment must be provided in a file named under 
        filename+"_dimensions". See data/experiments/illustrative_profile_dimensions
        for instructions.
        The default is True.

    Returns
    -------
    None.
    """
    x = np.linspace(0,1,num=301)
    meta = PCE() # used for inversion
    neld = Nelder_Mead(meta)
    exp_profile = u.load_profile(name=filename,path=path,do_nondim=do_nondim)
    couple = inverse(exp_profile,algo=neld)
    estim_profile = meta.evaluate(couple[0],couple[1])
    if plot :
        _ = plt.figure()
        plt.plot(x,exp_profile, label="Experiment")
        plt.plot(x,estim_profile, label="Metamodel + Nelder-Mead")
        plt.legend()
        plt.xlabel("x")
        plt.ylabel("h")
        plt.title("Comparison between experimental data and its fit by our procedure")
        plt.show()
    print(f"The estimation is B={couple[0]},S={couple[1]}")
