import numpy as np
from scipy.optimize import minimize

class Nelder_Mead :
    """
    Class wrapping the inversion procedure using Nelder-Mead. Use the inverse 
    method to solve an inverse problem.
    """
    
    def __init__(self, metamodel, maxiter=400, xtol=0.0001, ftol=0.00000001, bounds=((0.5,250.),(0.05,120.)),**useless):
        """
        Initialisation of the class.

        Parameters
        ----------
        metamodel : PCE
            The metamodel to use.
        maxiter : int, optional
            Maximal number of iter for Nelder-Mead. 
            The default is 400.
        xtol : float, optional
            Absolute error in xopt between iterations that is acceptable for convergence. 
            The default is 0.0001.
        ftol : float, optional
            Absolute error in error_func(xopt) between iterations that is acceptable for convergence. 
            The default is 0.00000001.
        bounds : sequence, optional
            Bounds on (B,S). 
            The default is ((0.5,250.),(0.05,120.)).
        **useless : 
            Container for useless parameters that could be passed to simplify an
            interface. NOT USED HERE.

        Returns
        -------
        None.
        """
        
        self.maxiter = maxiter
        self.xtol = xtol
        self.ftol = ftol
        self.bounds = bounds
        self.metamodel = metamodel
        self.guess = lambda x : np.array([125.,60.])
    
    def error_func(self, couple, profile, weights):
        """
        Error function used for the Nelder-Mead algorithm, which is the L2 norm
        of the difference between the profile and the result of the metamodel 
        for the given couple. Weights can be put on different portions of the
        profile.

        Parameters
        ----------
        couple : numpy array
            Couple B,S.
        profile : numpy array
            Height profile of the inverse problem.
        weights : numpy array
            Weights to be affected on the various profile components. Must be
            of shape 301.

        Returns
        -------
        float
            Error.
        """
        return np.linalg.norm(weights*(self.metamodel.evaluate(couple[0],couple[1])-profile))
    
    def inverse(self, profile, weights=np.ones(301)):
        """
        Solve the inverse problem (with Nelder-Mead method) for the given profile.   
        Weights can be put on different portions of the profile.

        Parameters
        ----------
        profile : numpy array
            Height profile of the inverse problem.
        weights : numpy array, optional
            Weights to be affected on the various profile components. Must be
            of shape 301. 
            The default is np.ones(301).

        Returns
        -------
        numpy array
            The (B,S) couple estimated by Nelder-Mead.
        """
        res = minimize(self.error_func, self.guess(profile), args=(profile, weights), method="Nelder-Mead", bounds=self.bounds, options={"maxiter" : self.maxiter, "xatol" : self.xtol, "fatol" : self.ftol})
        return res.x
