import os
import numpy as np
repo_path = os.environ["ZENODO_REPO_CPB"]

def scaling_from_dimensions(filename):
    """
    Compute the scaling parameter to use on a height profile to compute its
    nondimensional counterpart. The dimensions of the experiment must be given
    as shown in data/experiments/illustrative_profile_dimensions.

    Parameters
    ----------
    filename : str
        Name of the file containing the experiment dimensions.

    Returns
    -------
    float
        Scaling parameter for the height profile.
    """
    # loading dimensions
    dim = np.loadtxt(filename)
    g, L_scale, Q_scale, W, Qmh, rho, K, n, phiRad, Lmax = dim
    n = int(n)
    Lmax *= L_scale
    Qms = Qmh / Q_scale
    h0 = (K*Lmax/(rho*g*np.cos(phiRad))*(Qms/W)**n)**(1/(2*(n+1)))
    return L_scale/h0

def create_nondim_profile(filename):
    """
    Creates the nondimensional counterpart of a height profile contained in
    filename. The physical dimensions of the corresponding experiment are 
    assumed to be provided in a file named under filename+"_dimensions". See 
    data/experiments/illustrative_profile_dimensions for instructions. The
    new profile is saved under the name filename+"_nodim".

    Parameters
    ----------
    filename : str
        Name of the file containing the height profile.

    Returns
    -------
    None.
    """
    profile = np.loadtxt(filename)
    # loading scaling parameter
    scaling = scaling_from_dimensions(filename+"_dimensions")
    # modification of the curve
    profile *= scaling
    np.savetxt(filename+"_nodim", profile)

def load_profile(name,path=repo_path+"data/experiments/",do_nondim=True):
    """
    Loads a height profile in the folder path. Performs the nondimensionalisation
    or load it if available. It is possible to deactivate it and load an already
    nondimensional profile. To do the nondimensionalisation, the physical dimensions 
    of the corresponding experiment must be provided in a file named under 
    filename+"_dimensions". See data/experiments/illustrative_profile_dimensions
    for instructions.

    Parameters
    ----------
    name : str
        Name of the file.
    path : str, optional
        Path to the file. 
        The default is repo_path+"data/experiments/".
    do_nondim : bool, optional
        Whether to nondimensionalise or not. 
        The default is True.

    Returns
    -------
    numpy array
        The (possibly nondimensional) height profile.
    """
    if do_nondim :
        if not(name+"_nodim" in os.listdir(path)) :
            create_nondim_profile(path+name)
        name += "_nodim"
    return np.loadtxt(path+name)
