import os
import numpy as np
import matplotlib.pyplot as plt
import sklearn.decomposition as sk
repo_path = os.environ["ZENODO_REPO_CPB"]
os.chdir(repo_path)
from src.utils.data_base import heights, read_couple_grid

def PCA(Y=None, list_grid="all", n=1, error="last", output=True, plot=False, 
        log_scale=True, num_compo=6, title=True, n_components=301):
    """
    Performs a PCA on the data. Optionally plots the percentage of variance explained by each PC as
    well as some PCs (along the box).

    Parameters
    ----------
    Y : NoneType or numpy array, optional
        The list of heights. If None, they are extracted using "list_grid". 
        The default is None.
    list_grid : str or list of str, optional
        The grid(s) in which to take the heights. If "all", "n" should be set and all grids are 
        read for this n.
        The default is "all".
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. 
        The default is 1.
    error : str, optional
        Only the last error is implemented. 
        The default is "last".
    output : bool, optional
        If True, the PCA object is returned. 
        The default is True.
    plot : bool, optional
        If True, displays the diagnostics. 
        The default is False.
    log_scale : bool, optional
        If True, the explained variance graph is in log scale. 
        The default is True.
    num_compo : int, optional
        Number of components to display on the graph. 
        The default is 6.
    title : bool, optional
        It True, displays titles on the plots. 
        The default is True.
    n_components : int, optional
        Number of components to keep in the PCA. Must be higher than "num_compo". 
        The default is 301.

    Returns
    -------
    pca : scikit-learn PCA
        The PCA after fitting the data.
    """
    # extraction of the data
    if error != "last" :
        raise Exception("Not implemented yet")
    if Y.all() == None :
        print("Extraction of data")
        X = read_couple_grid(name=list_grid, n=n)
        Y = np.array([heights(B,S,1,steps=error) for B,S in X])
        print("Data extracted")
    # PCA
    pca = sk.PCA(n_components=n_components)
    _ = pca.fit(Y)
    # plot
    if plot :
        _ = plt.figure()
        plt.plot(100*pca.explained_variance_ratio_,"o")
        plt.xlabel("components")
        plt.ylabel("explained variance, %")
        if log_scale :
            plt.yscale("log")
            plt.xscale("log")
        if title :
            t = "Percentage of variance explained by each component of the PCA"
            if log_scale :
                t += ", in log scale"
        plt.show()
            
        _ = plt.figure()
        for i in range(num_compo) :
            plt.plot(np.linspace(0,1,num=301), pca.components_[i], label=str(i))
        plt.xlabel("x")
        plt.ylabel("PCA component")
        if title :
            plt.title("The {} first components of the PCA".format(num_compo))
        plt.legend()
        plt.show()
    if output :
        return pca
