import os
import numpy as np
import numpoly as npo
repo_path = os.environ["ZENODO_REPO_CPB"]
os.chdir(repo_path)
from src.metamodel.PCA import PCA

class PCE :
    """
    Class wrapping the PCE metamodel. Evaluation of the metamodel for one couple
    should be done using the evaluate method. For multiple couples, use
    evaluate_mult (more optimal than a loop).
    """
    
    def __init__(self, list_grid="regular_grid_1_n_1",n=1, out_feature="PCA",order=15,path=repo_path+"data/PCE_param/",n_compo=10,**useless):
        """
        Initialisation of the class. The files containing the PCE parameters are 
        assumed to be named using the convention defined in the function 
        train_on_grid of the module src.metamodel.PCE_trainer.

        Parameters
        ----------
        list_grid : str, optional
            The grid which has been used for training the PCE. 
            The default is "regular_grid_1_n_1".
        n : int or float, optional
            Value of the HB law power. If 1, must be an int. Useful only if there
            is an ambiguity in the name of the grid.
            The default is 1.
        out_feature : str, optional
            Whether to use the PCA components or the height profiles themselves for
            the PCE. The authorised values are "heights" and "PCA".
            The default is "PCA".
        order : int, optional
            Order of the PCE. 
            The default is 15.
        path : str, optional
            Path to the folder containing the PCE parameters. 
            The default is repo_path+"data/PCE_param/".
        n_compo : int, optional
            Number of components to use for the PCA. 
            The default is 10.
        **useless :
            Container for useless parameters that could be passed to simplify an
            interface. NOT USED HERE.

        Returns
        -------
        None.
        """
        
        print("Initialisation of the PCE model")
        self.list_grid = list_grid
        self.order = order
        self.out_feature = out_feature
        self.path = path
        if out_feature == "heights" :
            self.n_compo = 301
        else :
            self.n_compo = n_compo
        name = f"{list_grid}_{out_feature}_order_{order}"
        self.poly = npo.load(path+name)[:self.n_compo]
        print("Done, ready for evaluations")
        if out_feature=="PCA" :
            pca = PCA(Y=self.heights,num_compo=n_compo,n_components=n_compo)
            self.transform = lambda x: pca.inverse_transform(x)
        else :
            self.transform = lambda x: x
    
    def eval_compo(self, B, S, normalized=False):
        """
        Raw evaluation of the PCE, ignoring PCA or heights for the ouput.

        Parameters
        ----------
        B : float
            "rheology". See paper.
        S : float
            "slope". see paper.
        normalized : bool, optional
            If False, the couples are shifted and rescaled to belong to the domain
            [-1,1]x[-1,1]. It is assumed that the original (B,S) domain is 
            [0.5,250]x[0.05,120].
            The default is False.

        Returns
        -------
        numpy array
            Evaluation of the PCE.
        """
        if normalized :
            b,s = B,S
        else :
            b = (B-125.25)/125.25
            s = (S-60.025)/60.025
        return self.poly(b,s)
    
    def evaluate(self, B, S, normalized=False):
        """
        Complete evaluation of the metamodel for a given (B,S) couple, outputing a
        height profile even if the PCE is used with the PCA.

        Parameters
        ----------
        B : float
            "rheology". See paper.
        S : float
            "slope". See paper.
        normalized : bool, optional
            If False, the couple is shifted and rescaled to belong to the domain
            [-1,1]x[-1,1]. It is assumed that the original (B,S) domain is 
            [0.5,250]x[0.05,120].
            The default is False.

        Returns
        -------
        numpy array
            Evaluation of the metamodel.
        """
        return self.transform(self.eval_compo(B,S,normalized=normalized))
    
    def evaluate_mult(self, B, S, normalized=False):
        """
        Complete evaluation of the metamodel for an array of (B,S) couples.

        Parameters
        ----------
        B : numpy array
            List of B.
        S : numpy array
            List of S.
        normalized : bool, optional
            If False, the couples are shifted and rescaled to belong to the domain
            [-1,1]x[-1,1]. It is assumed that the original (B,S) domain is 
            [0.5,250]x[0.05,120].
            The default is False.

        Returns
        -------
        numpy array
            Evaluation of the metamodel.
        """
        if self.out_feature == "heights" :
            return np.transpose(self.evaluate(B, S, normalized=normalized))
        else :
            return self.transform(np.transpose(self.eval_compo(B,S,normalized=normalized)))
