import os
import numpy as np
import numpoly
import chaospy
repo_path = os.environ["ZENODO_REPO_CPB"]
os.chdir(repo_path)
import src.utils.data_base as d
from src.metamodel.PCA import PCA

def train_1D(couples, labels, order=4, save_name=None, output=False, force=False):
    """
    Computes a PCE of a scalar-valued function f(B,S) given a set of precomputed 
    evaluations of the aforementioned function.

    Parameters
    ----------
    couples : numpy array
        The (B,S) couples. The expected shape is (num_couples,2).
    labels : numpy array
        The corresponding f(B,S) of shape (num_couples).
    order : int, optional
        Order of the PCE. 
        The default is 4.
    save_name : NoneType or str, optional
        Name of the file in which to save the computed parameters. They are not
        saved if save_name=None. The file is necessarily saved in the folder
        data/PCE_param/.
        The default is None.
    output : bool, optional
        Whether to return the result of the PCE or not. 
        The default is False.
    force : bool, optional
        If set to False the function is only executed if there is not already
        a file under the name save_name in the folder data/PCE_param/. It is 
        ignored if save_name=None. 
        The default is False.

    Returns
    -------
    fitted_polynomial : chaospy.ndpoly
        The PCE.
    """
    if not(save_name==None) and not(force) and (save_name in os.listdir(repo_path+"data/PCE_param/")):
        print("previous training found, pass")
    else :
        q0, q1 = numpoly.variable(2)
        l = []
        for q in q0,q1 :
            lq = [1]
            lq.append(q)
            for n in range(1,order+1) :
                lq.append(((2*n+1)*q*lq[-1]-n*lq[-2])/(n+1))
            lq=numpoly.polynomial(lq)
            l.append(lq)
        P1 = l[0]
        P2 = l[1]
        P = []
        for i in range(order+1) :
            for j in range(order+1-i) :
                p = P1[i]*P2[j]
                if not(p in P) :
                    P.append(p)
        P = numpoly.polynomial(P)
        fitted_polynomial = chaospy.fit_regression(P, couples.T, labels)
        if not(save_name==None) :
            numpoly.save(f"{repo_path}data/PCE_param/{save_name}", fitted_polynomial)
            print("res saved")
        if output :
            return fitted_polynomial

def train_multiD(real_couples, labels, order=4, save_name=None, output=False, force=False, change_input=True):
    """
    Computes a PCE of a function f(B,S) with values in R^d given a set of precomputed 
    evaluations of the aforementioned function. It uses the previous train_1D function
    on all components of the given function.

    Parameters
    ----------
    real_couples : numpy array
        The (B,S) couples. The expected shape is (num_couples,2).
    labels : numpy array
        The corresponding f(B,S) of shape (num_couples,d).
    order : int, optional
        Order of the PCE. 
        The default is 4.
    save_name : NoneType or str, optional
        Name of the file in which to save the computed parameters. They are not
        saved if save_name=None. The file is necessarily saved in the folder
        data/PCE_param/.
        The default is None.
    output : bool, optional
        Whether to return the result of the PCE or not. 
        The default is False.
    force : bool, optional
        If set to False the function is only executed if there is not already
        a file under the name save_name in the folder data/PCE_param/. It is 
        ignored if save_name=None. 
        The default is False.
    change_input : bool, optional
        If True, the couples are shifted and rescaled to belong to the domain
        [-1,1]x[-1,1]. It is assumed that the original (B,S) domain is 
        [0.5,250]x[0.05,120]. Generally enhances performances. The default is True.

    Returns
    -------
    poly : numpoly.ndpoly
        The PCE.
    """
    if not(save_name==None) and not(force) and (save_name in os.listdir(repo_path+"data/PCE_param/")):
        print("previous training found, pass")
    else :
        if change_input :
            couples = np.zeros_like(real_couples)
            couples[:,0] = (real_couples[:,0]-125.25)/125.25
            couples[:,1] = (real_couples[:,1]-60.025)/60.025
        else :
            couples = real_couples
        poly = []
        print_every = labels.shape[1]/10
        for i in range(labels.shape[1]) :
            if i%print_every == 0 :
                print(f"{i}th component")
            p = train_1D(couples, labels[:,i], order=order, output=True)
            poly.append(p)
        poly = numpoly.polynomial(poly)
        if not(save_name==None) :
            numpoly.save(f"{repo_path}data/PCE_param/{save_name}", poly)
            print("res saved")
        if output :
            return poly

def train_on_grid(grid="regular_grid_1_n_1", order=15, label="PCA", force=False, n_compo=10, n=1):
    """
    Computes a PCE using a given grid (provided by the PDE solver). The user must
    choose whether to use a PCA as described in the paper or to use the 301 
    components of the height profiles.

    Parameters
    ----------
    grid : str, optional
        The grid to use for training (see src/utils/data_base.py for details). 
        The default is "regular_grid_1_n_1".
    order : int, optional
        Order of the PCE. 
        The default is 15.
    label : str, optional
        Whether to use the PCA components or the height profiles themselves for
        training. The authorised values are "heights" and "PCA".
        The default is "PCA".
    force : bool, optional
        If set to False the function is only executed if there is not already a
        file under the name f"{grid}_{label}_order_{order}" in the folder data/PCE_param/. 
        The default is False.
    n_compo : int, optional
        Number of components to use for the PCA. 
        The default is 10.
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. Useful only if there
        is an ambiguity in the name of the grid.
        The default is 1.

    Returns
    -------
    None.
    """
    print(f"grid {grid}, order {order}")
    print("loading data")
    real_couples = d.read_couple_grid(grid,n=n)
    couples = np.zeros_like(real_couples)
    couples[:,0] = (real_couples[:,0]-125.25)/125.25
    couples[:,1] = (real_couples[:,1]-60.025)/60.025
    heights = d.read_grid_heights(grid,n=n)
    if label=="heights" :
        labels = heights
        print("using heights")
        save_name=f"{grid}_{label}_order_{order}"
        train_multiD(couples, labels,order=order,save_name=save_name,force=force,change_input=False)
    elif label=="PCA" :
        print("using PCA")
        save_name=f"{grid}_{label}_order_{order}"
        pca = PCA(Y=heights,num_compo=n_compo,n_components=n_compo)
        labels = pca.transform(heights)
        print("PCE")
        train_multiD(couples, labels,order=order,save_name=save_name,force=force,change_input=False)
    else :
        print("label error, no training done")
