import os
import zipfile as z
import numpy as np
repo_path = os.environ["ZENODO_REPO_CPB"]


#### reading of solutions from data_base

def heights(B, S, n, steps="all"):
    """
    Extracts the height profiles for all snapshots at times "steps".

    Parameters
    ----------
    B : float
    S : float
    n : int or float
        Value of the HB law power. If 1, must be an int.
    steps : str or numpy array of int or list of int, optional
        The snapshots to extract. If "all", extracts all snapshots. If "last", extracts only the 
        wall touch. Otherwise, the list should contain number between 0 and 49. 
        The default is "all".

    Returns
    -------
    numpy array of shape (steps.shape, 301)
        The height profile for all snapshots.
    """
    file = repo_path+"data/PDE_solutions/n_{}_B_{}_S_{}".format(n, B, S)
    tab = np.loadtxt(file, skiprows=1)
    if steps == "last" :
        steps = -1
    elif steps == "all" :
        steps = np.arange(0, tab.shape[0])
    return tab[steps, :-4]

def durations(B,S,n):
    """
    Extracts the duration of computation of the full solution.

    Parameters
    ----------
    B : float
    S : float
    n : int or float
        Value of the HB law power. If 1, must be an int.

    Returns
    -------
    float
        The duration of computation in seconds.

    """
    file = repo_path+"data/PDE_solutions/n_{}_B_{}_S_{}".format(n, B, S)
    f = open(file,"r")
    duration = float(f.readline().split("duration ")[1].split()[0])
    f.close()
    return duration


#### managing descriptors of data_base

def extract_all_couples(n=1, path=repo_path+"data/PDE_solutions/", array=True):
    """
    Extract all the couples of the database corresponding to the given "n". Files should 
    have a name of the form :
        n_1_B_1._S_100.

    Parameters
    ----------
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. 
        The default is 1.
    path : str, optional
        The directory where the database is stored. 
        The default is "data/PDE_solutions/".
    array : bool, optional
        If True, returns a numpy array. Otherwise returns a native python list. 
        The default is True.

    Returns
    -------
    list or numpy array
        All couples.

    """
    # extracting files from the path
    res = []
    try :
        files = os.listdir(path)
    except FileNotFoundError :
        raise FileNotFoundError("The directory in which you are searching for data does not exist.")
    for file in files :
        # start with a "try" to protect against a non-compliant file.
        try :
            _, n1, _, B, _, S = file.split("_") # extraction of the parameters
            if float(n1) == n : # add only if it is the correct n
                # conversion of str to float
                B = float(B)
                S = float(S)
                # adaptation according to the system of coordinates
                res.append([B,S])
        except ValueError :
            print("The file {} is not conform, it is ignored.".format(file))
            pass
    if array :
        res = np.array(res)
    return res

def create_grid_list(name, header="Couples are of the form B,S with a space between each couple."):
    """
    Initialise a descriptor of a grid. The first line of the file is a description of the grid. The 
    second one is a list of tuples.

    Parameters
    ----------
    name : str
        Name of the file. Must finish with "n_" plus the actual value of n.
    header : str, optional
        A header to describe the grid. Will be the first line of the file. 
        The default is "Couples are of the form B,S with a space between each couple.".

    Returns
    -------
    None.

    """
    if header[-1] != "\n" : # to manage the separation of the two lines of the file
        header += "\n"
    f = open(repo_path+"data/data_base_descriptors/"+name, "w")
    f.write(header)
    f.close()
    print("Created file '{}' in '{}data/data_base_descriptors/'.".format(name,repo_path))
    
def list_grids(n=1):
    """
    Lists all the names of the grid descriptors for the given "n".

    Parameters
    ----------
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. 
        The default is 1.

    Returns
    -------
    res : list
        The list of names of grids.

    """
    file = repo_path+"data/data_base_descriptors/list_grids"
    f = open(file, "r")
    lines = f.readlines()
    f.close()
    pattern = "_n_"+str(n) # the pattern guarantees that the correct n is selected
    res = []
    for line in lines :
        if line[-1-len(pattern):-1] == pattern :
            res.append(line[:-1])
    return res

def read_couple_grid(name="regular_grid_1_n_1", n=1, array=True):
    """
    Returns the list of all couples of the given grid(s). Even if name is set to "all", only the couples
    that are registered in a descriptor will be considered.

    Parameters
    ----------
    name : str or list of str, optional
        The name(s) of the grid(s). If "all", all grids corresponding to the given "n" are read. 
        The default is "regular_grid_1_n_1".
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. Used only if "name" is set to "all".
        The default is 1.
    array : bool, optional
        If True, returns a numpy array. Otherwise returns a native python list. 
        The default is True.

    Returns
    -------
    res : list or numpy array
        All the couples.

    """
    # determine the list of grids
    if name == "all" :
        name = list_grids(n=n)
    elif type(name) == str :
        name = [name]
    res = []
    for nam in name : # iterates over the different grids
        file = repo_path+"data/data_base_descriptors/"+nam
        try : # protection from an error in the name of a grid
            f = open(file, "r")
            line = f.readlines()
            if len(line) == 1: # this means that the descriptor is empty
                line = []
            else :
                line = line[1].split() # separates the couples
            f.close()
        except FileNotFoundError :
            raise FileNotFoundError("The name of the grid is not correct. "+
                                    "Call 'list_grids(n)' for the list possibilities "+
                                        "or create a grid first using 'create_grid_list'.")
        for couple in line :
            # turns each str into list of float
            B, S = couple.split(",")
            B = float(B)
            S = float(S)
            # check for the system of coordinates
            res.append([B, S])
    if array :
        res = np.array(res)
    return res

def update_MC_couples_grid(n=1, grids_ignore="all", couples_ignore=[]):
    """
    Update the descriptor of the Monte Carlo sampling, after the sampling. It takes all couples of the
    database and add them to the descriptor if they do not correspond to another descriptor. Rewrite 
    completely the file each time.

    Parameters
    ----------
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. 
        The default is 1.
    grids_ignore : str or list of str, optional
        The grids that are supposed not to intersect with the MC sampling. If set to "all", all 
        grids are considered. 
        The default is "all".
    couples_ignore : list of list, optional
        List of couples to specifically ignore even if they are not in a grid. 
        The default is [].

    Returns
    -------
    None.

    """
    print("Loading couples to ignore")
    if grids_ignore != "all" and type(grids_ignore) != list :
        raise Exception("The parameter 'grids_ignore' takes either a list or the value 'all'.")
    prefix = repo_path+"data/data_base_descriptors/"
    # starting with the specific given couples
    ignore = couples_ignore
    # adding couples of the grids to ignore
    if grids_ignore == "all" :
        grids_ignore = list_grids(n=n)
    for name in grids_ignore :
        if name != "monte_carlo_n_{}".format(n) :
            ignore += read_couple_grid(name=name, array=False)
    print("Updating the descriptor")
    file = prefix + "monte_carlo_n_{}".format(n)
    c = extract_all_couples(n=n)
    f = open(file,"w")
    f.write("The grid given by a Monte Carlo sampling."+ \
            " Couples are of the form B,S with a space between each couple.\n")
    for B, S in c :
        if not([B,S] in ignore):
            f.write(str(B)+","+str(S)+" ")
    f.close()

def update_couples_grid(BS, name, check=True):
    """
    Update a grid descriptor with a list of couples. The descriptor is assumed to exist.

    Parameters
    ----------
    BS : list or numpy array
        List of couples.
    name : str
        Name of the descriptor.
    check : bool, optional
        If True, checks for redundancies when updating. 
        The default is True.

    Returns
    -------
    None.

    """
    prefix = repo_path+"data/data_base_descriptors/"
    file = prefix + name
    ignore = read_couple_grid(name=name, array=False)
    f = open(file,"a")
    for B, S in BS :
        if not(check) or not([B,S] in ignore):
            f.write(str(B)+","+str(S)+" ")
    f.close()


def create_grid_heights(list_grid="regular_grid_1_n_1", n=1):
    """
    Create a file in the archive "data_base_per_grid" which contains the heights for all the couples 
    of the given grids (for quicker reading).

    Parameters
    ----------
    list_grid : str of list of str, optional
        The grid or the list of the grids. Do not use "all" here. 
        The default is "regular_grid_1_n_1".
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. 
        The default is 1.

    Returns
    -------
    None.

    """
    name = list_grid
    print("Creating grid heights under name :",name)
    couples = read_couple_grid(list_grid,n=n)
    h = np.zeros((couples.shape[0],50,301))
    for i in range(couples.shape[0]) :
        h[i] = heights(couples[i,0],couples[i,1],n)
    np.save(repo_path+"data/tmp.npy", h)
    f = z.ZipFile(repo_path+"data/data_base_per_grid.zip", "a", z.ZIP_DEFLATED)
    f.write(repo_path+"data/tmp.npy", name)
    f.close()
    os.remove(repo_path+"data/tmp.npy")

def read_grid_heights(list_grid="regular_grid_1_n_1", n=1, steps="last"):
    """
    Read heights from a file in "data_base_per_grid".

    Parameters
    ----------
    list_grid : str of list of str, optional
        The grid or the list of the grids. Do not use "all" here. 
        The default is "regular_grid_1_n_1".
    n : int or float, optional
        Value of the HB law power. If 1, must be an int. 
        The default is 1.
    steps : str or numpy array of int or list of int, optional
        The snapshots to extract. If "all", extracts all snapshots. If "last", extracts only the wall 
        touch. Otherwise, the list should contain numbers between 0 and 49. 
        The default is "last".

    Returns
    -------
    numpy array of shape (n_couples, steps.shape, 301)
        The height profiles for all snapshots.

    """
    name = list_grid
    try :
        f = z.ZipFile(repo_path+"data/data_base_per_grid.zip", "r")
        h = np.load(f.open(name,"r"))
    except :
        print("First call to this grid, please wait for the creation of a shortcut")
        create_grid_heights(list_grid=list_grid,n=n)
        f = z.ZipFile(repo_path+"data/data_base_per_grid.zip", "r")
        h = np.load(f.open(name,"r"))
    f.close()
    if steps == "last" :
        steps = -1
    elif steps == "all" :
        steps = np.arange(0, h[0].shape[0])
    return h[:,steps]
